/*
 * Some tests for assignment 2 CO7104 at the University of Leicester 2013/14
 * Copyright (c) 2013, Lorenz (LAHS1)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <string>

#include "Sparse.hpp"


using std::cout;
using std::cin;
using std::endl;
using std::flush;
using std::string;

int _numDigits(int v) {
	int res(0);
	if (v < 0) {
		res++;
		v *= -1;
	}
	res += ((v >= 1000000000) ? 10 : (v >= 100000000) ? 9 : (v >= 10000000) ? 8
		: (v >= 1000000) ? 7 : (v >= 100000) ? 6 : (v >= 10000) ? 5
		: (v >= 1000) ? 4 : (v >= 100) ? 3 : (v >= 10) ? 2 : 1);
	return res;
}

/* The abort message would be nicer if this were a macro, but macros aren't
 * particularly nice and the use of a debugger is recommended anyway if
 * something fails */
void assertString(const Sparse& sparse, const string& expected, const string& error) {
	if(sparse.toString() != expected) {
		cout << "ERROR: Matrix output differs from expected result."
		<< " Expected output:" << endl << expected << "Actual output:" << endl << sparse.toString();
		assert(false);
	}
}

// Some basic tests
void testBasics() {
	int width = 4;
	int height = 2;
	int** arr = new int*[width];
	for(int i = 0; i < width; ++i) {
		arr[i] = new int[height];
		for (int j = 0; j < height; ++j) {
			arr[i][j] = i+j*width;
		}
	}
	Sparse sparse(width, height, arr);
	assertString(sparse, "0 1 2 3\n4 5 6 7\n",
		"Array constructer produced incorrect result");

	assert(!sparse.insert(3,2,9));
	assertString(sparse, "0 1 2 3\n4 5 6 7\n      9\n",
		"Inserting elements does not work");
	assert(!sparse.insert(1,2,7));
	assert(sparse.insert(1,1,8));
	assertString(sparse, "0 1 2 3\n4 8 6 7\n  7   9\n",
		"Inserting elements does not work reliably");
	assert(!sparse.insert(2,3,1));
	assert(!sparse.insert(3,3,2));
	assert(!sparse.insert(2,2,8));
	assertString(sparse, "0 1 2 3\n4 8 6 7\n  7 8 9\n    1 2\n",
		"Inserting elements does not work reliably");

	// Add an element that will increase column width
	assert(!sparse.insert(0,3,10));
	assertString(sparse, " 0  1  2  3\n 4  8  6  7\n    7  8  9\n10     1  2\n",
		"Column width is most likely wrong");

	// Further increase column width
	assert(!sparse.insert(5,1,105));
	assertString(sparse, "  0   1   2   3\n  4   8   6   7     105\n      7   8   9\n 10       1   2\n",
		"Column width is most likely wrong");

	// Add an array in position (1,2)
	assert(sparse.addArrayOverwrite(width, height, arr, 1, 2));   // should work
	assert(!sparse.addArrayOverwrite(width, height, arr, 4, 1));  // should fail, position does not exist
	assertString(sparse, "  0   1   2   3\n  4   8   6   7     105\n      0   1   2   3\n 10   4   5   6   7\n",
		"Adding of array does not work");

	// Reduce column width again
	assert(sparse.insert(5,1,11));
	string temp = " 0  1  2  3\n 4  8  6  7    11\n    0  1  2  3\n10  4  5  6  7\n";
	if (sparse.toString() != temp) {
		cout << "WARNING: Column width should be reduced again! Expected"
			<< endl << temp << "but got" << endl << sparse.toString()
			<< "SUBSEQUENT TEST CASES MAY ASSUME THIS BEHAVIOUR, SO TEST "
			<< "FAILURES SHOULD BE INVESTIGATED CAREFULLY, MAYBE THE PROBLEM "
			<< "IS ONLY toString BEHAVIOUR" << endl;
	}

	// add a sparse matrix in position (5,1)
	Sparse copy(4,2,arr);
	copy.insert(3,2,9);
	copy.insert(1,2,7);
	copy.insert(1,1,2);
	copy.insert(2,3,1);
	copy.insert(3,3,2);
	copy.insert(2,2,8);
	copy.insert(0,3,5);
	copy.insert(5,1,3);
	copy.addArrayOverwrite(4, 2, arr, 1, 2);
	assertString(copy, "0 1 2 3\n4 2 6 7   3\n  0 1 2 3\n5 4 5 6 7\n",
		"Matrix construction is not realiable");
	sparse.addSparseOverwrite(copy, 5, 1);
	assertString(sparse, " 0  1  2  3\n 4  8  6  7     0  1  2  3\n    0  1  2  3  4  2  6  7     3\n10  4  5  6  7     0  1  2  3\n                5  4  5  6  7\n",
		"Adding of sparse matrix failed");

	for (int i = 0; i < width; ++i) {
		delete[] arr[i];
	}
	delete[] arr;

	// test 'value' function
	int val(2);
	assert(sparse.value(0,0,val) && val == 0);
	val = 12345;
	assert(!sparse.value(0,2,val) && val == 12345);
	assert(sparse.value(10,2,val) && val == 3);
	val = 0;
	assert(sparse.value(9,3,val) && val == 3);
	val = 23456;
	assert(!sparse.value(9,1,val) && val == 23456);
	assert(sparse.value(5,1,val) && val == 0);
	assert(sparse.value(0,3,val) && val == 10);
}

// test addArrayPush{Down,Right}
void testArrayPush() {
	// 200    120
	//     21  33
	//  97 45
	Sparse sparse;
	sparse.insert(0,0,200);
	sparse.insert(2,0,120);
	sparse.insert(1,1,21);
	sparse.insert(2,1,33);
	sparse.insert(0,2,97);
	sparse.insert(1,2,45);
	assertString(sparse, "200     120\n     21  33\n 97  45\n",
		"Construction of sparse matrix failed");
	// 350 384
	//   0  25
	int** arr = new int*[2];
	arr[0] = new int[2];
	arr[1] = new int[2];
	arr[0][0] = 350; arr[0][1] = 0; arr[1][0] = 384; arr[1][1] = 25;
	// do a pushDown
	sparse.addArrayPushDown(2,2,arr,1,1);
	assertString(sparse, "200     120\n    350 384\n      0  25\n     21  33\n 97  45\n",
		"addArrayPushDown failed");

	// add the same array with pushRight
	sparse.addArrayPushRight(2,2,arr,2,2);
	assertString(sparse, "200             120\n    350         384\n      0 350 384  25\n     21   0  25  33\n 97  45\n",
		"addArrayPushRight failed");
	delete[] arr[0];
	delete[] arr[1];
	delete[] arr;
}

// test adding elements that are above or to the left of the current top-left element
void testAddTopLeftLater() {
	Sparse sparse;
	sparse.insert(1,1,5);
	assertString(sparse, "\n  5\n",
		"Case where top-left element is not (0,0) is not handled correctly");

	// add row above topleft
	sparse.insert(1,0,3);
	assertString(sparse, "  3\n  5\n",
		"Case where row is inserted *above* previous top-left element is not handled correctly");

	// add cell left of topleft in same row
	sparse.insert(0,0,4);
	assertString(sparse, "4 3\n  5\n",
		"Case where element is inserted to the left of previous top-left element is not handled correctly");
}

// test inserting elements into various gaps
void testInsertIntoGap() {
	Sparse sparse;
	sparse.insert(1,1,2);
	sparse.insert(5,1,6);
	sparse.insert(0,2,3);
	sparse.insert(2,2,5);
	sparse.insert(1,5,1);
	sparse.insert(4,6,4);
	assertString(sparse, "\n  2       6\n3   5\n\n\n  1\n        4\n",
		"Construction of matrix with lots of gaps failed");

	// test the 'exists' function
	assert(sparse.exists(1,1));
	assert(sparse.exists(0,2));
	assert(sparse.exists(4,6));

	assert(!sparse.exists(2,3));
	assert(!sparse.exists(100000,100000));
	assert(!sparse.exists(3,3));
	assert(!sparse.exists(0-2,0-2));  // 4294967294

	// insert stuff into various gaps
	sparse.insert(2,0,1);
	sparse.insert(2,2,7);
	sparse.insert(4,1,8);
	sparse.insert(5,2,9);
	sparse.insert(3,1,1);
	sparse.insert(2,4,5);
	sparse.insert(4,4,8);
	sparse.insert(3,5,9);
	sparse.insert(0,5,3);
	sparse.insert(2,6,1);
	sparse.insert(3,6,2);
	sparse.insert(0,6,5);
	sparse.insert(6,3,8);
	assertString(sparse, "    1\n  2   1 8 6\n3   7     9\n            8\n    5   8\n3 1   9\n5   1 2 4\n",
		"Failed to insert elements into various gaps");
}

// test support for elements with value 0
void testZeroElements() {
	int width = 3;
	int height = 3;
	int** arr = new int*[height];
	for(int i = 0; i < height; ++i) {
		arr[i] = new int[width];
		for (int j = 0; j < width; ++j) {
			arr[i][j] = i+width*j;
		}
	}
	Sparse sparse(width, height, arr);
	assertString(sparse, "0 1 2\n3 4 5\n6 7 8\n",
		"Construction of matrix from 3x3 array failed");

	assert(sparse.insert(1,1,0));
	assert(sparse.insert(1,2,0));
	assert(sparse.insert(1,0,0));
	assertString(sparse, "0 0 2\n3 0 5\n6 0 8\n",
		"Setting of element to 0 failed");

	arr[2][2] = 0;
	sparse.addArrayOverwrite(width, height, arr, 1, 2);
	assertString(sparse, "0 0 2\n3 0 5\n6 0 1 2\n  3 4 5\n  6 7 0\n",
		"Insertion of array at position with value 0 failed");

	for (int i = 0; i < height; ++i) {
		delete[] arr[i];
	}
	delete[] arr;
}

void testNavigation() {
	// build matrix
	Sparse sparse;
	sparse.insert(1,0,1);
	sparse.insert(2,0,2);
	sparse.insert(4,0,3);
	sparse.insert(0,1,4);
	sparse.insert(2,1,5);
	sparse.insert(3,1,6);
	sparse.insert(4,2,7);
	sparse.insert(5,2,8);
	sparse.insert(1,3,9);
	sparse.insert(3,3,0);
	sparse.insert(4,3,1);
	sparse.insert(0,4,2);
	sparse.insert(1,4,3);
	sparse.insert(2,4,4);
	sparse.insert(5,4,5);
	// Check if it was constructed correctly
	assertString(sparse, "  1 2   3\n4   5 6\n        7 8\n  9   0 1\n2 3 4     5\n",
		"Matrix was constructed incorrectly");

	unsigned int x(1), y(0);
	assert(sparse.nextRow(x,y) && y == 3);
	assert(sparse.nextCol(x,y) && x == 3);
	assert(sparse.prevRow(x,y) && y == 1);
	assert(!sparse.nextCol(x,y) && x == 3);
	assert(!sparse.prevRow(x,y) && y == 1);
	assert(sparse.prevCol(x,y) && x == 2);
	assert(sparse.prevRow(x,y) && y == 0);
	assert(sparse.nextRow(x,y) && y == 1);
	assert(sparse.prevCol(x,y) && x == 0);
	assert(sparse.nextRow(x,y) && y == 4);
	assert(sparse.nextCol(x,y) && x == 1);
	assert(sparse.nextCol(x,y) && x == 2);
	assert(sparse.nextCol(x,y) && x == 5);
	assert(!sparse.nextCol(x,y) && x == 5);
	assert(!sparse.nextRow(x,y) && y == 4);
	assert(sparse.prevRow(x,y) && y == 2);  // going up when cell->up() is NULL
	assert(!sparse.prevRow(x,y) && y == 2);
	assert(sparse.prevCol(x,y) && x == 4);
	assert(!sparse.prevCol(x,y) && x == 4);
	x = 3; y = 1;
	assert(sparse.nextRow(x,y) && y == 3);  // going down when cell->down() is NULL
	x = 0; y = 1;
	assert(!sparse.prevCol(x,y) && !sparse.prevRow(x,y) && x == 0 && y == 1);
}

void testNavigationForum() {
	// Test navigation with the example from the forum
	// https://campus.cs.le.ac.uk/teaching/forums/viewtopic.php?f=9&t=22
	Sparse sparse;
	sparse.insert(0,0,0);
	sparse.insert(1,0,1);
	sparse.insert(2,0,2);
	sparse.insert(0,1,3);
	sparse.insert(2,1,4);
	sparse.insert(0,2,5);
	unsigned int y(0);
	assert(!sparse.nextRow(1,y) && y == 0);
	sparse.insert(1,2,6);
	sparse.insert(2,2,7);
	assert(sparse.nextRow(1,y) && y == 2);
}


void _createDiagonalMatrix(Sparse& sparse, int size, int diagonalWidth) {
	for (int j = -diagonalWidth+1; j < diagonalWidth; ++j) {
		for (int i = 0; i < size; ++i) {
			if (i+j < 0 || i+j >= size) {
				continue;
			}
			sparse.insert(i + j, i, diagonalWidth - (j < 0 ? -j : j));
		}
	}
}


void _testBigDiagonal(unsigned int size, unsigned int diagonalWidth) {
	Sparse sparse;
	int val(-1), diff(-1);
	unsigned int x(0), y(0);
	_createDiagonalMatrix(sparse, size, diagonalWidth);
	for (unsigned int i = 0; i < size; ++i) {
		for (unsigned int j = 0; j < size; ++j) {
			val = -1;
			diff = static_cast<int>(i) - static_cast<int>(j);
			if (diff < 0) {
				diff *= -1;
			}
			x = i; y = j;
			if (static_cast<unsigned int>(diff) < diagonalWidth) {
				assert(sparse.value(i, j, val) && val == static_cast<int>(diagonalWidth)-diff);

				if (i-j == -diagonalWidth+1) {
					// left end of the diagonal
					assert(!sparse.prevCol(x,y) && x == i);
					assert(!sparse.nextRow(x,y) && y == j);
					if (i < size-1)
						assert(sparse.nextCol(x,y) && x == i+1);
					if (j > 0)
						assert(sparse.prevRow(i,y) && y == j-1);
				} else if (i-j == diagonalWidth-1) {
					// right end of the diagonal
					assert(!sparse.nextCol(x,y) && x == i);
					assert(!sparse.prevRow(x,y) && y == j);
					if (i > 0)
						assert(sparse.prevCol(x,y) && x == i-1);
					if (j < size-1)
						assert(sparse.nextRow(i,y) && y == j+1);
				} else {
					// somewhere in the middle of the diagonal
					if (i < size-1)
						assert(sparse.nextCol(x,j) && x == i+1);
					else
						assert(!sparse.nextCol(x,j) && x == i);

					if (j < size - 1)
						assert(sparse.nextRow(i,y) && y == j+1);
					else
						assert(!sparse.nextRow(i,y) && y == j);

					x = i; y = j;
					if (x > 0)
						assert(sparse.prevCol(x,j) && x == i-1);
					else
						assert(!sparse.prevCol(x,j) && x == i);

					if (y > 0)
						assert(sparse.prevRow(i,y) && y == j-1);
					else
						assert(!sparse.prevRow(i,y) && y == j);
				}
			} else {
				// not set
				assert(!sparse.value(i, j, val) && val == -1);
				assert(!sparse.prevCol(x,y) && x == i);
				assert(!sparse.nextCol(x,y) && x == i);
				assert(!sparse.prevRow(x,y) && y == j);
				assert(!sparse.nextCol(x,y) && y == j);
			}
		}
	}
}


void testBigDiagonal(unsigned int maxsize = 100, unsigned int factor = 10) {
	Sparse sparse;
	unsigned int size(10);
	unsigned int diagonalWidth(2);
	_createDiagonalMatrix(sparse, size, diagonalWidth);
	assertString(sparse, "2 1\n1 2 1\n  1 2 1\n    1 2 1\n      1 2 1\n        1 2 1\n          1 2 1\n            1 2 1\n              1 2 1\n                1 2\n",
		"Creation of 10x10 diagonal matrix failed");

	while (size <= maxsize) {
		// diagonal = 2^(log_10(size)+2)
		int diagonal = 1;
		for (int i = 0; i <= _numDigits(size); ++i) {
			diagonal *= 2;
		}
		cout << "Testing size " << size << " with diagonal " << diagonal << "... " << flush;

		_testBigDiagonal(size, diagonal);
		cout << "done" << endl;

		if (size < maxsize && size * factor > maxsize) {
			size = maxsize;
		} else {
			size *= factor;
		}
	}
}


void testExampleFromMail() {
	int col1[4] = {0, 1, 2, 3};
	int col2[4] = {4, 5, 6, 7};
	int* rows[2] = {col1, col2};

	Sparse b(2,4,rows);
	assert(b.toString() == "0 4\n1 5\n2 6\n3 7\n");

	for (unsigned int i = 0 ; i < 2 ; ++i) {
		for (unsigned int j = 0 ; j < 4 ; ++j) {
			int value(0);
			b.value(i, j, value);
			assert(value == rows[i][j]);
		}
	}
}

void testNegativeValues() {
	int col1[4] = {0,-1,2,-3};
	int col2[4] = {4,5,-6,7};
	int* rows[2] = {col1,col2};

	Sparse sparse(2,4,rows);
	assertString(sparse, " 0  4\n-1  5\n 2 -6\n-3  7\n",
		"Negative values are handled incorrectly");

	assert(!sparse.insert(2,1,100));
	assert(!sparse.insert(2,3,-20));
	assertString(sparse, "  0   4\n -1   5 100\n  2  -6\n -3   7 -20\n",
		"Negative values are handled inconsistently");

	assert(!sparse.insert(2,2,-12345));
	assertString(sparse, "     0      4\n    -1      5    100\n     2     -6 -12345\n    -3      7    -20\n",
		"Column width not updated correctly for negative values");

	// Column width example from forum
	int forum_col1[2] = {100000,1};
	int forum_col2[2] = {1,2};
	int forum_col3[2] = {2,3};
	int* forum_rows[3] = {forum_col1, forum_col2, forum_col3};
	Sparse forum(3, 2, forum_rows);
	assertString(forum, "100000      1      2\n     1      2      3\n",
		"Column width incorrect");
}

int main(int argc, char** argv) {
	testBasics();
	testArrayPush();
	testAddTopLeftLater();
	testInsertIntoGap();
	testZeroElements();
	testNavigation();
	testNavigationForum();
	testExampleFromMail();
	testNegativeValues();

	// set up big diagonal test
	int maxsize(100);
	int step(10);

	if (argc > 1) {
		maxsize = atoi(argv[1]);
	}
	if (argc > 2) {
		step = atoi(argv[2]);
	}
	testBigDiagonal(maxsize, step);

	cout << "Passed all tests" << endl;
	return 0;
}
